use crate::radio::MAX_PACKET_LEN;
use anyhow::anyhow;
use colored::{Color, Colorize};
use ham_cats::{
    packet::Packet,
    whisker::{Route, RouteIdentity, RouteNode},
};

pub fn print_packet(pkt: &Packet<MAX_PACKET_LEN>, rssi: Option<f64>) {
    println!();

    let color = match rssi {
        Some(rssi) => {
            println!("----- RF [{rssi} dBm] -----");

            Color::Green
        }
        None => {
            println!("--- FELINET ---");

            Color::Magenta
        }
    };

    if let Some(ident) = pkt.identification() {
        println!(
            "{} \t{}-{} [ICON: {}]",
            "IDENT:".color(color),
            ident.callsign,
            ident.ssid,
            ident.icon
        );
    }
    if let Some(route) = pkt.route() {
        let r_str = route
            .iter()
            .map(|rn| match rn {
                ham_cats::whisker::RouteNode::Internet => "[NET]".to_string(),
                ham_cats::whisker::RouteNode::Identity(ident) => {
                    let s = if ident.is_future() { "*" } else { "" };
                    let rc = ident.callsign();
                    let rs = ident.ssid();
                    let rssi = match ident.rssi() {
                        Some(x) => format!("[{x:.1} dbm]"),
                        None => "".to_string(),
                    };
                    format!("{rc}-{rs}{s} {rssi}")
                }
            })
            .collect::<Vec<String>>()
            .join(" -> ");
        println!(
            "{} \t(MAX {}) {r_str}",
            "ROUTE:".color(color),
            route.max_hops
        );
    }

    let mut buf = [0; MAX_PACKET_LEN];
    if let Ok(comment) = pkt.comment(&mut buf) {
        println!("{} \t'{comment}'", "CMNT:".color(color));
    }
    if let Some(gps) = pkt.gps() {
        println!(
            "{} \t({:07.2}, {:07.2}) +/- {:03} m, v = {:.2} m/s [N {:.2} deg]",
            "GPS:".color(color),
            gps.latitude(),
            gps.longitude(),
            gps.max_error,
            gps.speed,
            gps.heading()
        );
    }

    if let Some(info) = pkt.node_info() {
        match (info.hardware_id(), info.software_id()) {
            (Some(hw), Some(sw)) => {
                println!(
                    "{} \t{hw:#06x} {} {sw:#04x}",
                    "HW:".color(color),
                    "SW:".color(color)
                );
            }
            (Some(hw), None) => {
                println!("{} \t{hw:#06x}", "SW:".color(color));
            }
            (None, Some(sw)) => {
                println!("{} \t{sw:#04x}", "SW:".color(color));
            }
            (None, None) => {}
        }

        if let Some(uptime) = info.uptime() {
            println!("{} \t{} s", "UTIME:".color(color), uptime);
        }

        if let Some(a_height) = info.antenna_height() {
            println!("{} \t{} m", "VERT:".color(color), a_height);
        }

        if let Some(a_gain) = info.antenna_gain() {
            println!("{} \t{} dBi", "GAIN:".color(color), a_gain);
        }

        if let Some(power) = info.tx_power() {
            println!("{} \t{} dBm", "TXP:".color(color), power);
        }

        if let Some(volts) = info.voltage() {
            println!("{} \t{} V", "VOLTS:".color(color), volts);
        }

        if let Some(temp) = info.xcvr_temperature() {
            println!("{} \t{} C", "TEMP:".color(color), temp);
        }
    }
}

pub fn attempt_decode<'a>(
    packet: &[u8],
    buf: &'a mut [u8; MAX_PACKET_LEN],
    rssi: f64,
) -> anyhow::Result<Packet<'a, MAX_PACKET_LEN>> {
    let packet: Packet<MAX_PACKET_LEN> =
        Packet::fully_decode(packet, buf).map_err(|e| anyhow!("Could not decode packet: {e:?}"))?;

    print_packet(&packet, Some(rssi));

    Ok(packet)
}

// i-gating is kind of weird.
// if a packet has a pre-determined route (is_future=true) and our node
// doesn't match it, we don't digipeat it.
// But FELINET should be an "all-seeing eye". Any packet that's heard
// should end up on the network.
// The solution is that, if the packet has non-matching is_future=true nodes,
// We rip them off, and stick our node + an internet route on
#[must_use]
pub fn append_internet_to_packet_route(
    callsign: &str,
    ssid: u8,
    rssi: f64,
    packet: &mut Packet<MAX_PACKET_LEN>,
) -> Option<()> {
    let mut route = packet.route().unwrap_or(Route::new(0));

    let first_future = route.iter().find_map(|rn| match rn {
        RouteNode::Identity(ident) if ident.is_future() => Some((ident.callsign(), ident.ssid())),
        _ => None,
    });

    if let Some((rc, rs)) = first_future {
        if rc == callsign && rs == ssid {
            // we can just do the normal thing
            route.append_node(callsign, ssid, Some(rssi)).ok()?;
        } else {
            // rip off all the futures and add ourselves
            let mut new_route = Route::new(route.max_hops);
            for r in route.iter() {
                match r {
                    RouteNode::Internet => new_route.push_internet()?,
                    RouteNode::Identity(ident) => {
                        if ident.is_future() {
                            break;
                        }

                        new_route.push_callsign(ident);
                    }
                }
            }

            route = new_route;
            route.push_callsign(RouteIdentity::new(callsign, ssid, Some(rssi), false));
        }
    } else {
        route.append_node(callsign, ssid, Some(rssi)).ok()?;
    }

    route.push_internet()?;
    packet.clear_route();
    packet.add_route(route).ok()?;

    Some(())
}
